"""LLM-related utility functions."""

from typing import List, Optional

from langchain.chat_models.base import BaseChatModel
from langchain.llms.base import BaseLLM
from langchain.schema import BaseLanguageModel, PromptValue


def call_llm(
    llm: BaseLanguageModel, prompt: PromptValue, stop: Optional[List[str]] = None
) -> str:
    """Invoke the LLM's __call__ function.

    The BaseLanguageModel class does not have that function because the function
    signatures are slightly different for each child class, but you can readily get a
    str back all the same.
    """
    if isinstance(llm, BaseLLM):
        return llm(prompt=prompt.to_string(), stop=stop)
    if isinstance(llm, BaseChatModel):
        return llm(messages=prompt.to_messages(), stop=stop).content
    raise ValueError(f"Unknown type of LLM: {llm}")



"""Module for safe invocation of langchain models."""
from typing import Dict, TypeVar, Union

from langchain.chains.base import Chain
from langchain.prompts.base import BasePromptTemplate

T = TypeVar("T")


def safe_inputs(
    lc_object: Union[Chain, BasePromptTemplate], inputs: Dict[str, T]
) -> Dict[str, T]:
    """Filter for the subset of inputs that correspond to the given object."""
    if isinstance(lc_object, Chain):
        input_variables = lc_object.input_keys
    elif isinstance(lc_object, BasePromptTemplate):
        input_variables = lc_object.input_variables
    else:
        raise ValueError(f"Unknown type for {lc_object}")
    return {k: v for k, v in inputs.items() if k in input_variables}



"""Chain for selecting a MRKL tool for use."""
from __future__ import annotations

import json
import re
from typing import Dict, List, Tuple, Type

from langchain.chains.base import Chain
from langchain.llms.base import BaseLLM
from langchain.prompts.base import BasePromptTemplate
from langchain.schema import BaseLanguageModel
from langchain.tools.base import BaseTool

from langchain_contrib.utils import call_llm

from .prompt import MrklPromptSelector

FINAL_ANSWER_ACTION = "Final Answer:"


class MrklPickActionChain(Chain):
    """MRKL chain that selects a tool for use."""

    llm: BaseLanguageModel
    """LLM to run this iteration of the MRKL agent with."""
    prompt: BasePromptTemplate
    """Prompt for a single response of the LLM."""
    observation_prefix: str = "Observation: "
    """Observation prefix to stop the MRKL at."""
    choice_key: str = "choice"
    """Output key for which action the LLM chose."""
    action_input_key: str = "action_input"
    """Output key for the action's input."""

    @classmethod
    def from_tools(
        cls,
        llm: BaseLanguageModel,
        tools: List[BaseTool],
        prompt_selector: Type[MrklPromptSelector] = MrklPromptSelector,
    ) -> MrklPickActionChain:
        """Instantiate the MRKL action chain from a list of tools."""
        prompt = prompt_selector.from_tools(tools=tools).get_prompt(llm)
        return cls(llm=llm, prompt=prompt)

    @property
    def input_keys(self) -> List[str]:
        """Input keys this chain expects."""
        return self.prompt.input_variables

    @property
    def output_keys(self) -> List[str]:
        """Output keys this chain expects."""
        return ["observation", self.choice_key, self.action_input_key]

    def get_action_and_input(self, llm_output: str) -> Tuple[str, str, str]:
        """Parse out the action and input from the LLM output.

        Copied and edited from langchain/agents/mrkl/base.py
        """
        if FINAL_ANSWER_ACTION in llm_output:
            return "", "Final Answer", llm_output.split(FINAL_ANSWER_ACTION)[-1].strip()
        regex = r"(.*)[\n]*Action: (.*?)[\n]*Action Input: (.*)"
        match = re.search(regex, llm_output, re.DOTALL)
        if not match:
            raise ValueError(f"Could not parse LLM output: `{llm_output}`")
        observation = match.group(1).strip()
        action = match.group(2).strip()
        action_input = match.group(3)
        return observation, action, action_input.strip(" ").strip('"')

    def get_chat_action_and_input(self, llm_output: str) -> Tuple[str, str, str]:
        """Parse out the action and input from the LLM output.

        Copied and edited from langchain/agents/chat/base.py
        """
        if FINAL_ANSWER_ACTION in llm_output:
            return "", "Final Answer", llm_output.split(FINAL_ANSWER_ACTION)[-1].strip()
        try:
            match = re.search(r"Thought:(.*)\n*Action:", llm_output, re.DOTALL)
            assert match is not None
            observation = match.group(1).strip()
            _, action, _ = llm_output.split("```")
            response = json.loads(action.strip())
            return observation, response["action"], response["action_input"]

        except Exception:
            raise ValueError(f"Could not parse LLM output: {llm_output}")

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        """Get the LLM to pick an action and its input."""
        llm_response = call_llm(
            self.llm,
            self.prompt.format_prompt(**inputs),
            stop=[f"\n{self.observation_prefix.rstrip()}"],
        )
        if isinstance(self.llm, BaseLLM):
            observation, action, input = self.get_action_and_input(llm_response)
        else:
            observation, action, input = self.get_chat_action_and_input(llm_response)
        return {
            "observation": observation,
            self.choice_key: action,
            self.action_input_key: input,
        }



"""Module to reimplement langchain's MRKL in a more customizable manner."""

from .choice import MrklLoopChain
from .pick_action import MrklPickActionChain

__all__ = [
    "MrklPickActionChain",
    "MrklLoopChain",
]



"""Patch for Terminal class."""

from typing import Callable

from vcr.cassette import Cassette
from vcr_langchain.patch import GenericPatch, add_patchers

from ..terminal import Terminal


class TerminalPatch(GenericPatch):
    """Patch for Terminal class."""

    def __init__(self, cassette: Cassette) -> None:
        """Initialize patch."""
        super().__init__(cassette, Terminal, "_get_raw_shell_update")

    def get_same_signature_override(self) -> Callable:
        """Obtain same-signature override for Terminal._get_raw_shell_update."""

        def _call(og_self: Terminal, cmd: str) -> str:
            return self.generic_override(og_self, cmd=cmd)

        return _call


add_patchers(TerminalPatch)



"""Types useful for prompting."""

from typing import Union

from langchain.prompts import PromptTemplate
from langchain.prompts.base import BasePromptTemplate
from langchain.prompts.chat import BaseMessagePromptTemplate, ChatPromptTemplate
from langchain.schema import BaseMessage

Templatable = Union[str, BaseMessagePromptTemplate, BaseMessage, BasePromptTemplate]
"""Anything that can be converted directly into a BasePromptTemplate."""


def into_template(templatable: Templatable) -> BasePromptTemplate:
    """Convert a Templatable into a proper BasePromptTemplate."""
    if isinstance(templatable, str):
        return PromptTemplate.from_template(templatable)
    elif isinstance(templatable, BaseMessagePromptTemplate) or isinstance(
        templatable, BaseMessage
    ):
        return ChatPromptTemplate.from_messages([templatable])
    elif isinstance(templatable, BasePromptTemplate):
        return templatable
    else:
        raise ValueError(f"Don't know how to convert {templatable} into template")



"""Prompts and templates for choices."""

from .prompt_value import (
    BaseChoicePrompt,
    ChatChoicePrompt,
    ChoiceStr,
    StringChoicePrompt,
)
from .template import (
    ChoicePromptTemplate,
    get_oxford_comma_formatter,
    get_simple_joiner,
    list_of_choices,
)

__all__ = [
    "BaseChoicePrompt",
    "StringChoicePrompt",
    "ChatChoicePrompt",
    "ChoiceStr",
    "ChoicePromptTemplate",
    "get_simple_joiner",
    "get_oxford_comma_formatter",
    "list_of_choices",
]



"""Experimental LLM chains."""

from .chained import ChainedPromptTemplate
from .choice import ChoicePromptTemplate
from .prefixed import PrefixedTemplate
from .schema import Templatable, into_template

__all__ = [
    "ChainedPromptTemplate",
    "PrefixedTemplate",
    "ChoicePromptTemplate",
    "Templatable",
    "into_template",
]



"""Module that defines the choice prompt."""
from __future__ import annotations

from typing import Any, Callable, List, Sequence, Union

from langchain.prompts.base import BasePromptTemplate
from langchain.prompts.chat import BaseMessagePromptTemplate, ChatPromptTemplate
from langchain.prompts.prompt import PromptTemplate
from langchain.schema import BaseMessage
from pydantic import Field

from langchain_contrib.utils import f_join

from .prompt_value import BaseChoicePrompt

ChoicesFormatter = Callable[[List[str]], str]


def get_simple_joiner(joiner: str = ", ") -> ChoicesFormatter:
    """Get a choice formatter that's just a simple joining of strings."""

    def simple_join_choice(choices: List[str]) -> str:
        """Do a simple join on the choice strings."""
        return f_join(joiner, choices)

    return simple_join_choice


def get_oxford_comma_formatter(conjunction: str = "or") -> ChoicesFormatter:
    """Get a choice formatter that respects the Oxford comma."""

    def oxford_comma_list(choices: List[str]) -> str:
        """Phrase the list using the Oxford comma."""
        if len(choices) == 0:
            return ""
        elif len(choices) == 1:
            return choices[0]
        elif len(choices) == 2:
            return f_join(f" {conjunction} ", choices)
        else:
            head = f_join(", ", choices[:-1])
            return f_join(f", {conjunction} ", [head, choices[-1]])

    return oxford_comma_list


def list_of_choices(choices: List[str]) -> str:
    """Return a numerical list of choices."""
    return f_join("\n", [f"{i+1}. {choice}" for i, choice in enumerate(choices)])


class ChoicePromptTemplate(BasePromptTemplate):
    """A wrapper prompt template for picking from a number of choices.

    This template preserves choice information in prompts.
    """

    base_template: BasePromptTemplate
    """The base template that this class wraps around."""
    choices: List[str]
    """The list of choices to pick from."""
    choices_formatter: ChoicesFormatter = Field(
        default_factory=get_oxford_comma_formatter
    )
    """How to convert from the list of choices to a single string.

    Utility functions to help with this include:

    - get_simple_joiner
    - get_oxford_comma_formatter
    - list_of_choices
    """
    choice_format_key: str = "choices"
    """Which string is used for formatting choices in the template."""

    @classmethod
    def from_base_template(
        cls, base_template: BasePromptTemplate, choices: List[str], **kwargs: Any
    ) -> ChoicePromptTemplate:
        """Load a ChoicePromptTemplate from base templates."""
        result = cls(
            base_template=base_template,
            input_variables=base_template.input_variables,
            choices=choices,
            **kwargs,
        )

        result.partial_variables = {
            **result.partial_variables,
            result.choice_format_key: result._no_args_formatter,
        }
        result.input_variables.remove(result.choice_format_key)
        return result

    @classmethod
    def from_template(
        cls, template: str, choices: List[str], **kwargs: Any
    ) -> ChoicePromptTemplate:
        """Load a ChoicePromptTemplate from a text template."""
        base_template = PromptTemplate.from_template(template)
        return cls.from_base_template(
            base_template=base_template, choices=choices, **kwargs
        )

    @classmethod
    def from_messages(
        cls,
        messages: Sequence[Union[BaseMessagePromptTemplate, BaseMessage]],
        choices: List[str],
        **kwargs: Any,
    ) -> ChoicePromptTemplate:
        """Load a ChoicePromptTemplate from message templates."""
        base_template = ChatPromptTemplate.from_messages(messages)
        return cls.from_base_template(
            base_template=base_template, choices=choices, **kwargs
        )

    @property
    def _prompt_type(self) -> str:
        return "choice"

    def _no_args_formatter(self) -> str:
        """No-argument choice formatter for partials."""
        return self.choices_formatter(self.choices)

    def format(self, **kwargs: Any) -> str:
        """Format the prompt with the inputs."""
        return self.format_prompt(**kwargs).to_string()

    def format_prompt(self, **kwargs: Any) -> BaseChoicePrompt:
        """Format the prompt while preserving the choices."""
        kwargs = self._merge_partial_and_user_variables(**kwargs)
        prompt = self.base_template.format_prompt(**kwargs)
        return BaseChoicePrompt.from_prompt(prompt, choices=self.choices)



"""Defines the Chained prompt template type."""

from typing import Any, List

from langchain.prompts.base import (
    BasePromptTemplate,
    StringPromptTemplate,
    StringPromptValue,
)
from langchain.prompts.chat import ChatPromptValue
from langchain.schema import PromptValue

from langchain_contrib.utils import f_join, safe_inputs

from .schema import Templatable, into_template


class ChainedPromptTemplate(StringPromptTemplate):
    """A prompt template composed of multiple other prompt templates chained together.

    This is a StringPromptTemplate rather than a BasePromptTemplate to enable use in
    BaseStringMessagePromptTemplate.
    """

    joiner: str = ""
    """How to join each template output together.

    Only meaningful for StringPromptTemplate's.
    """
    subprompts: List[BasePromptTemplate]

    def __init__(self, subprompts: List[Templatable], joiner: str = "", **kwargs: Any):
        """Initialize a ChainedPromptTemplate.

        subprompts can be passed in as just plain strings for convenience.
        """
        prompts = [into_template(p) for p in subprompts]
        input_variables = list(
            set([var for subprompt in prompts for var in subprompt.input_variables])
        )
        super().__init__(
            input_variables=input_variables,
            joiner=joiner,  # type: ignore
            subprompts=prompts,  # type: ignore
            **kwargs,
        )

    def format(self, **kwargs: Any) -> str:
        """Format the prompt with the inputs."""
        return self.format_prompt(**kwargs).to_string()

    def format_prompt(self, **kwargs: Any) -> PromptValue:
        """Format each series of prompts with the given inputs."""
        unused_args = set(kwargs.keys())
        values: List[PromptValue] = []
        for subprompt in self.subprompts:
            prompt_args = safe_inputs(subprompt, kwargs)
            unused_args -= prompt_args.keys()
            values.append(subprompt.format_prompt(**prompt_args))
        if unused_args:
            raise KeyError(unused_args)

        all_strings = all([isinstance(x, StringPromptValue) for x in values])
        if all_strings:
            text = f_join(self.joiner, [x.to_string() for x in values])
            return StringPromptValue(text=text)
        else:
            messages = [
                message for subvalue in values for message in subvalue.to_messages()
            ]
            return ChatPromptValue(messages=messages)

    @property
    def _prompt_type(self) -> str:
        """Return the prompt type key."""
        return "chained"



"""Terminal with persistent shell between commands."""

from . import patchers  # noqa: F401
from .safety import SafeTerminalChain, TerminalToolChain
from .terminal import Terminal
from .tool import TerminalTool

__all__ = [
    "Terminal",
    "TerminalTool",
    "TerminalToolChain",
    "SafeTerminalChain",
]



"""Patch for Terminal class."""

from typing import Callable, List, Optional

from vcr.cassette import Cassette
from vcr_langchain.patch import GenericPatch, add_patchers

from .llm import Human


class HumanPatch(GenericPatch):
    """Patch for the Human class."""

    def __init__(self, cassette: Cassette) -> None:
        """Initialize patch."""
        super().__init__(cassette, Human, "_call")

    def get_same_signature_override(self) -> Callable:
        """Obtain same-signature override for Human._call."""

        def _call(og_self: Human, prompt: str, stop: Optional[List[str]] = None) -> str:
            return self.generic_override(og_self, prompt=prompt, stop=stop)

        return _call


add_patchers(HumanPatch)



"""Module for the human LLM."""

from .llm import BaseHuman, Human

try:
    from . import patchers  # noqa: F401
except ImportError:
    pass  # vcr_langchain is an optional dependency

__all__ = [
    "BaseHuman",
    "Human",
]



"""Module that defines a Chain wrapper for Tools."""

from typing import Dict, List

from langchain.chains.base import Chain
from langchain.tools.base import BaseTool


class ToolChain(Chain):
    """Wraps a Tool in a Chain.

    Allows for tool and chain interop outside of the Agent class.
    """

    tool: BaseTool
    """The tool this chain will call when invoked."""
    tool_input_key: str = "action_input"
    """The key for this class to feed into the tool."""
    tool_output_key: str = "action_result"
    """The key produced by this class as a result of using the tool."""

    @property
    def input_keys(self) -> List[str]:
        """Input keys this chain expects."""
        return [self.tool_input_key]

    @property
    def output_keys(self) -> List[str]:
        """Output keys this chain expects."""
        return [self.tool_output_key]

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        """Run the tool for this chain."""
        tool_input = inputs[self.tool_input_key]
        tool_output = self.tool.run(tool_input)
        return {self.tool_output_key: tool_output}



"""Prompting configuration for MRKL agents."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List

from fvalues import F
from langchain.chains.prompt_selector import ConditionalPromptSelector, is_chat_model
from langchain.prompts.base import BasePromptTemplate
from langchain.prompts.chat import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)
from langchain.prompts.prompt import PromptTemplate
from langchain.tools.base import BaseTool
from pydantic import BaseModel, Extra

from langchain_contrib.prompts import ChoicePromptTemplate
from langchain_contrib.prompts.choice import get_simple_joiner
from langchain_contrib.utils import f_join


class BaseMrklPrompt(BaseModel, ABC):
    """A base class for MRKL prompting."""

    tools: List[BaseTool]
    """Tools that the MRKL agent has access to."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    @property
    @abstractmethod
    def base_prompt(self) -> BasePromptTemplate:
        """The basis for the MRKL prompt."""

    @property
    def tool_names(self) -> List[str]:
        """Just the names of the tools, without their descriptions."""
        return [tool.name for tool in self.tools]

    @property
    def tool_descriptions(self) -> F:
        """Descriptions of the tools and their usage."""
        return f_join(
            "\n", [F(f"{tool.name}: {tool.description}") for tool in self.tools]
        )

    def template(self) -> ChoicePromptTemplate:
        """Return a ChoicePromptTemplate for these tools."""
        partial_choice = ChoicePromptTemplate.from_base_template(
            base_template=self.base_prompt,
            choices=self.tool_names,
            choice_format_key="tool_names",
            choices_formatter=get_simple_joiner(),
        ).partial(
            tool_descriptions=self.tool_descriptions,
        )
        assert isinstance(partial_choice, ChoicePromptTemplate)
        return partial_choice


class StringMrklPrompt(BaseMrklPrompt):
    """The string version of the prompt to the MRKL agent."""

    template_text: str = """
Answer the following questions as best you can. You have access to the following tools:

{tool_descriptions}

Use the following format:

Question: the input question you must answer
Thought: you should always think about what to do
Action: the action to take, should be one of [{tool_names}]
Action Input: the input to the action
Observation: the result of the action
... (this Thought/Action/Action Input/Observation can repeat N times)
Thought: I now know the final answer
Final Answer: the final answer to the original input question

Begin!

Question: {input}
Thought:{agent_scratchpad}""".lstrip()  # noqa

    @property
    def base_prompt(self) -> BasePromptTemplate:
        """The basis for the string MRKL prompt."""
        return PromptTemplate.from_template(self.template_text)


class ChatMrklPrompt(BaseMrklPrompt):
    """The chat version of the prompt to the MRKL agent."""

    system_template: str = """
Answer the following questions as best you can. You have access to the following tools:

{tool_descriptions}

The way you use the tools is by specifying a json blob.
Specifically, this json should have a `action` key (with the name of the tool to use) and a `action_input` key (with the input to the tool going here).

The only values that should be in the "action" field are: {tool_names}

The $JSON_BLOB should only contain a SINGLE action, do NOT return a list of multiple actions. Here is an example of a valid $JSON_BLOB:

```
{{{{
  "action": $TOOL_NAME,
  "action_input": $INPUT
}}}}
```

Here is an example of an invalid $JSON_BLOB:

```
{{{{
  "action": $FIRST_TOOL_NAME,
  "action_input": $FIRST_INPUT
}}}}

{{{{
  "action": $SECOND_TOOL_NAME,
  "action_input": $SECOND_INPUT
}}}}
```

ALWAYS use the following format:

Question: the input question you must answer
Thought: you should always think about what to do
Action:
```
$JSON_BLOB
```
Observation: the result of the action
... (this Thought/Action/Observation can repeat N times)
Thought: I now know the final answer
Final Answer: the final answer to the original input question

Begin! Reminder to always use the exact characters `Final Answer` when responding.
""".strip()  # noqa

    human_template: str = """{input}\n\n{agent_scratchpad}"""

    @property
    def base_prompt(self) -> BasePromptTemplate:
        """The basis for the chat MRKL prompt."""
        return ChatPromptTemplate.from_messages(
            messages=[
                SystemMessagePromptTemplate.from_template(self.system_template),
                HumanMessagePromptTemplate.from_template(self.human_template),
            ]
        )


class MrklPromptSelector(ConditionalPromptSelector):
    """Prompt definitions for the MRKL agent."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    @classmethod
    def from_tools(cls, tools: List[BaseTool]) -> MrklPromptSelector:
        """Construct MRKL prompt selector from a list of tools."""
        string_prompt_template = StringMrklPrompt(tools=tools).template()
        chat_prompt_template = ChatMrklPrompt(tools=tools).template()
        return cls(
            default_prompt=string_prompt_template,
            conditionals=[(is_chat_model, chat_prompt_template)],
        )



"""Module for code that patches various utilities to work with Terminal."""

from . import langchain_patch  # noqa: F401

try:
    from . import vcr_patch  # noqa: F401
except ImportError:
    pass  # vcr_langchain is an optional dependency



"""Utility code meant for development rather than agents."""

from .contexts import current_directory, temporary_file
from .fvalues import f_join
from .llm import call_llm
from .safe import safe_inputs

__all__ = [
    "current_directory",
    "temporary_file",
    "f_join",
    "safe_inputs",
    "call_llm",
]



"""Module to interact with a virtual terminal."""

import os
import shlex
import time
from typing import Any

import pexpect
from pydantic import BaseModel

from .ansi_escapes import remove_ansi_escapes


class UnknownResult(Exception):
    """Exception raised when terminal output is not as expected."""


class Terminal(BaseModel):
    """A virtual terminal that supports interactive shell commands.

    This will actually change the current program's working directory in response to
    `cd` commands. This allows the LLM to subsequently edit files according to their
    relative paths in the new directory.
    """

    shell: pexpect.spawn
    """The actual shell we're interacting with."""
    bash_prompt: str
    """The prompt of the shell that tells us the last command has finished running."""
    refresh_interval: float
    """How frequently we should check for shell updates."""
    output_size: int
    """The maximum number of shell output characters to read at once."""

    class Config:
        """pydantic config object."""

        arbitrary_types_allowed = True

    def __init__(
        self,
        bash_prompt: str = "zamm$ ",
        refresh_interval: float = 0.1,
        init_delay: float = 0.1,
        output_size: int = 1000,
        **kwargs: Any,
    ) -> None:
        """Initialize a virtual terminal.

        Args:
            refresh_interval: How long to wait between terminal reads.
            init_delay: How long to wait for initial terminal prompt during init.
            output_size: How many characters to read at a time.
            bash_prompt: Constant Bash prompt to use for terminal.
            **kwargs: Additional arguments to pass to `BaseModel`.
        """
        os.environ["PS1"] = bash_prompt
        sh = pexpect.spawn("/bin/bash --norc", encoding="utf-8")
        time.sleep(init_delay)
        bash_prompt = sh.read_nonblocking(size=output_size)
        super().__init__(
            refresh_interval=refresh_interval,
            output_size=output_size,
            shell=sh,
            bash_prompt=bash_prompt,
            **kwargs,
        )

    @property
    def prompt_length(self) -> int:
        """Get the length of the terminal prompt."""
        return len(self.bash_prompt)

    def _get_raw_shell_update_uncached(self, cmd: str) -> str:
        """Get the raw terminal output for a command."""
        self.shell.sendline(cmd)
        results = ""
        try:
            while not results.endswith(self.bash_prompt):
                latest_output = self.shell.read_nonblocking(size=self.output_size)
                results += latest_output
                time.sleep(0.1)
        except pexpect.TIMEOUT as e:
            raise UnknownResult(
                "Terminal output does not have initial prompt of: "
                f"'{self.bash_prompt}':\n\n{results}"
            ) from e

        # todo: more robust way of syncing terminal actions to action chain state
        parsed_cmd = shlex.split(cmd)
        if len(parsed_cmd) == 2 and parsed_cmd[0] == "cd":
            os.chdir(parsed_cmd[1])

        return results

    def _get_raw_shell_update(self, cmd: str) -> str:
        """Get the raw terminal output for a command.

        Call this function for terminal commands that can/should be cached.
        """
        return self._get_raw_shell_update_uncached(cmd)

    def _is_terminal_state_command(self, cmd: str) -> bool:
        """Check if a command changes the terminal state."""
        return cmd.startswith("cd ")

    def _get_shell_update(self, cmd: str) -> str:
        """Get the raw terminal output for a command.

        Results may or may not be cached, depending on whether the command is known to
        modify shell state.
        """
        if self._is_terminal_state_command(cmd):
            return self._get_raw_shell_update_uncached(cmd)
        return self._get_raw_shell_update(cmd)

    def run_bash_command(self, cmd: str) -> str:
        """Run a command in the terminal.

        Returns the interpreted output of the command as a single string.

        Args:
            cmd: The command to run.

        Raises:
            UnknownResult: If the terminal output does not end with the
                expected prompt.
        """
        results = self._get_shell_update(cmd)
        assert results.startswith(cmd), (
            f"'{results}' does not start with '{cmd}'. "
            "Is non-ASCII terminal input involved?"
        )
        unix_results = results.replace("\r\n", "\n")
        # + 1 to remove leading "\n" after command input
        output = unix_results[len(cmd) + 1 : -self.prompt_length]
        without_ansi = remove_ansi_escapes(output)
        return without_ansi



"""Fake LLMs for testing purposes."""
from typing import Any, List, Mapping, Optional

from langchain.llms.base import LLM
from pydantic import BaseModel


class FakeLLM(LLM, BaseModel):
    """Fake LLM wrapper for testing purposes."""

    mapped_responses: Optional[Mapping] = None
    """Mapping of prompt to predetermined responses."""
    sequenced_responses: Optional[List[str]] = None
    """List of responses to return in order.

    Useful if the prompt is too complicated to reconstruct for testing.
    """
    num_calls: int = 0
    """Keeps track of which sequenced response to return."""
    check_stops: bool = False
    """Set to true to check that stops are being set, and being set properly.
    `queries` must be modified to have the stop present.
    """

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "fake"

    def _check_stops(self, result: str, stop: Optional[List[str]]) -> str:
        if not self.check_stops:
            return result

        assert stop is not None, "Stop has not been set"
        found_stop = False
        for s in stop:
            if result.endswith(s):
                found_stop = True
                result = result[: len(result) - len(s)]
                break
        assert found_stop, f"Output '{result}' does not end in {stop}"
        return result

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        """First try to lookup in queries, else return 'foo' or 'bar'."""
        if self.sequenced_responses is not None and self.num_calls < len(
            self.sequenced_responses
        ):
            result = self.sequenced_responses[self.num_calls]
        elif self.mapped_responses is not None and prompt in self.mapped_responses:
            result = self.mapped_responses[prompt]
        else:
            result = self._uncached_queries(prompt, stop)

        self.num_calls += 1
        return self._check_stops(result, stop)

    def _uncached_queries(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        """Logic for queries that aren't hardcoded ahead of time."""
        if stop is None:
            return "foo"
        else:
            return "bar"

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        return {}



"""Chain that chooses and performs the next action."""
from abc import ABC
from typing import Callable, Dict, List, Mapping

from langchain.chains.base import Chain
from pydantic import BaseModel

from langchain_contrib.utils import safe_inputs


class ChoiceChain(Chain, BaseModel, ABC):
    """Chain that asks the LLM for a decision and executes it."""

    choice_picker: Chain
    """The chain that actually prompts the LLM for the choice."""
    prep_picker_output: Callable[[Dict[str, str]], Dict[str, str]] = lambda x: x
    """Interprets output from the picker chain for the chosen chain.

    Override this to do additional dict munging before it gets passed through to
    the chosen chain.
    """
    choices: Mapping[str, Chain]
    """The chains that will be run depending on the LLM's choice.

    This is a mapping from which LLM output corresponds to which chain.
    """
    choice_key: str = "choice"
    """choice_picker output key that tells us which choice was picked."""
    ignore_keys: List[str] = []
    """Keys that will be returned in final output, but not passed on to chosen chain."""

    @property
    def input_keys(self) -> List[str]:
        """Input keys to this chain."""
        return self.choice_picker.input_keys

    @property
    def output_keys(self) -> List[str]:
        """Possible output keys produced by this chain."""
        all_keys = set(self.choice_picker.output_keys)
        for choice in self.choices.values():
            all_keys.update(choice.output_keys)
        return list(all_keys)

    def _validate_outputs(self, outputs: Dict[str, str]) -> None:
        """Skip validation because different options may produce different outputs.

        It is assumed that each individual choice chain will validate its own output.
        """

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        """Run the logic of this chain and return the output."""
        raw_picker_output = self.choice_picker(inputs, return_only_outputs=True)
        ignored_output = {
            k: v for k, v in raw_picker_output.items() if k in self.ignore_keys
        }
        real_raw_output = {
            k: v for k, v in raw_picker_output.items() if k not in self.ignore_keys
        }

        picker_output = self.prep_picker_output(real_raw_output)
        if self.choice_key not in picker_output:
            raise KeyError(f"Choice-picking chain did not emit '{self.choice_key}'")
        choice = picker_output.pop(self.choice_key)
        assert isinstance(choice, str), f"Choice '{choice}' is not a str"
        if choice not in self.choices:
            raise KeyError(f"Choice picked does not exist: '{choice}'")
        chosen_chain = self.choices[choice]
        unused_args = picker_output.keys() - chosen_chain.input_keys
        if unused_args:
            raise KeyError(f"Extra input keys for choice: {unused_args}")

        full_inputs = {**inputs, **picker_output}
        chain_inputs = safe_inputs(chosen_chain, full_inputs)
        chain_output = chosen_chain(chain_inputs)
        return {
            self.choice_key: choice,
            **ignored_output,
            **picker_output,
            **chain_output,
        }



"""Fake chains for testing purposes."""

from typing import Callable, Dict, List, Optional

from langchain.chains.base import Chain
from pydantic import Extra


class FakeChain(Chain):
    """Fake chain that returns predefined outputs."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    expected_inputs: List[str] = []
    """List of input keys to expect."""
    expected_outputs: List[str] = []
    """List of output keys to expect.

    Not needed if `output` is defined.
    """
    output: Dict[str, str] = {}
    """The dict to return when this chain is called.

    This is ignored if `inputs_to_outputs` is defined.
    """
    inputs_to_outputs: Optional[Callable[[Dict[str, str]], Dict[str, str]]] = None
    """Function to transform inputs to outputs."""

    @property
    def input_keys(self) -> List[str]:
        """Input keys this chain expects."""
        return self.expected_inputs

    @property
    def output_keys(self) -> List[str]:
        """The keys of the predefined output dict."""
        return list(set(self.expected_outputs).union(self.output.keys()))

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        """Return the output dict, along with inputs if so specified."""
        if self.inputs_to_outputs is None:
            return self.output
        return self.inputs_to_outputs(inputs)



"""Module defining prompts involving choices."""

from __future__ import annotations

from typing import List

from fvalues import F
from langchain.prompts.base import StringPromptValue
from langchain.prompts.chat import ChatPromptValue
from langchain.schema import BaseMessage, PromptValue
from pydantic import Extra


class ChoiceStr(F):
    """String that keeps track of choices used to create this string."""

    choices: List[str]

    def __new__(cls, f_str: F, choices: List[str]) -> ChoiceStr:
        """Create a new ChoiceStr with cached choices."""
        result = super().__new__(cls, f_str, parts=f_str.parts)
        result.choices = choices
        return result


class BaseChoicePrompt(PromptValue):
    """A prompt that involves picking from a number of choices.

    This is just a wrapper around a regular PromptValue that preserves the choice
    information.
    """

    prompt: PromptValue
    """The encapsulated prompt that provides the actual prompt value."""
    choices: List[str]
    """The list of choices to choose from."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    def to_string(self) -> str:
        """Return prompt as string."""
        return self.prompt.to_string()

    def to_messages(self) -> List[BaseMessage]:
        """Return prompt as messages."""
        return self.prompt.to_messages()

    def _wrap_choice_str(self, str_prompt: str) -> ChoiceStr:
        """Wrap a prompt in ChoiceStr before returning."""
        if isinstance(str_prompt, F):
            f_prompt = str_prompt
        else:
            f_prompt = F(str_prompt)
        return ChoiceStr(f_prompt, self.choices)

    @classmethod
    def from_prompt(cls, prompt: PromptValue, choices: List[str]) -> BaseChoicePrompt:
        """Create one of the child classes from a base prompt."""
        if isinstance(prompt, StringPromptValue):
            return StringChoicePrompt(
                text=prompt.to_string(), prompt=prompt, choices=choices
            )
        elif isinstance(prompt, ChatPromptValue):
            return ChatChoicePrompt(
                messages=prompt.to_messages(), prompt=prompt, choices=choices
            )
        else:
            return cls(prompt=prompt, choices=choices)


class StringChoicePrompt(BaseChoicePrompt, StringPromptValue):
    """A string prompt that involves picking from a number of choices."""

    def to_string(self) -> str:
        """Return prompt as string."""
        str_prompt = self.prompt.to_string()
        return self._wrap_choice_str(str_prompt)

    def to_messages(self) -> List[BaseMessage]:
        """Return prompt as messages."""
        return self.prompt.to_messages()


class ChatChoicePrompt(BaseChoicePrompt, ChatPromptValue):
    """A chat prompt that involves picking from a number of choices."""

    def to_string(self) -> str:
        """Return prompt as string."""
        str_prompt = self.prompt.to_string()
        return self._wrap_choice_str(str_prompt)

    def to_messages(self) -> List[BaseMessage]:
        """Return prompt as messages."""
        return self.prompt.to_messages()



"""Module that adds safety to the terminal."""

from typing import Dict, List

from langchain.chains.base import Chain
from langchain.chains.llm import LLMChain
from langchain.chains.sequential import SequentialChain
from langchain.prompts.base import BasePromptTemplate
from langchain.prompts.prompt import PromptTemplate
from langchain.schema import BaseLanguageModel
from langchain.tools.base import BaseTool
from pydantic import Field

from langchain_contrib.chains import ChoiceChain, ToolChain
from langchain_contrib.llms import Human
from langchain_contrib.prompts import ChoicePromptTemplate

from .tool import TerminalTool


class TerminalToolChain(ToolChain):
    """Chain for running the Terminal."""

    tool: BaseTool = Field(default_factory=TerminalTool)
    tool_input_key: str = "command"
    tool_output_key: str = "output"


class SafeTerminalChain(Chain):
    """Chain that lets the human review terminal output before execution."""

    human: BaseLanguageModel = Field(default_factory=Human)
    """The LLM that will review this command for safety."""
    terminal: BaseTool = Field(default_factory=TerminalTool)
    """The terminal that this will use after human review."""
    proceed_choice: str = "Proceed"
    edit_command_choice: str = "Edit command"

    @property
    def input_keys(self) -> List[str]:
        """Input keys this chain expects."""
        return ["command"]

    @property
    def output_keys(self) -> List[str]:
        """Output keys this chain expects."""
        return ["command", "output", "choice"]

    @property
    def review_prompt(self) -> BasePromptTemplate:
        """Prompt the user for decision making."""
        return ChoicePromptTemplate.from_template(
            """
The LLM would like to run the command `{command}`. You can choose to {choices}.

Your choice: """.lstrip(),
            choices=[self.proceed_choice, self.edit_command_choice],
        )

    @property
    def edit_command_prompt(self) -> BasePromptTemplate:
        """Prompt the user for the new command."""
        return PromptTemplate.from_template("Replace it with: ")

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        pick_action = LLMChain(
            llm=self.human, prompt=self.review_prompt, output_key="choice"
        )
        run_terminal = TerminalToolChain(tool=self.terminal)
        edit_command = LLMChain(
            llm=self.human, prompt=self.edit_command_prompt, output_key="command"
        )
        review_chain = ChoiceChain(
            choice_picker=pick_action,
            choices={
                self.proceed_choice: run_terminal,
                self.edit_command_choice: SequentialChain(
                    chains=[edit_command, run_terminal],
                    input_variables=[],
                    output_variables=["command", "output"],
                    return_all=True,
                ),
            },
        )
        return review_chain(inputs, return_only_outputs=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['langchain_contrib',
 'langchain_contrib.chains',
 'langchain_contrib.chains.mrkl',
 'langchain_contrib.llms',
 'langchain_contrib.llms.human',
 'langchain_contrib.prompts',
 'langchain_contrib.prompts.choice',
 'langchain_contrib.tools',
 'langchain_contrib.tools.terminal',
 'langchain_contrib.tools.terminal.patchers',
 'langchain_contrib.utils']

package_data = \
{'': ['*']}

install_requires = \
['fvalues>=0.0.3,<0.0.4',
 'langchain>=0.0.119,<0.0.120',
 'pexpect>=4.8.0,<5.0.0',
 'simple-term-menu>=1.6.1,<2.0.0']

setup_kwargs = {
    'name': 'langchain-contrib',
    'version': '0.0.4',
    'description': '',
    'long_description': '# langchain-contrib\n\nA collection of utilities that are too experimental for [langchain proper](https://github.com/hwchase17/langchain), but are nonetheless generic enough to potentially be useful for multiple projects. Currently consists of code dumped from [ZAMM](https://github.com/amosjyng/zamm), but is of course open to contributions with lax procedures.\n\n## Quickstart\n\n```bash\npip install langchain-contrib\n```\n\nTo add interop with [`vcr-langchain`](https://github.com/amosjyng/vcr-langchain), simply install it as well:\n\n```bash\npip install vcr-langchain\n```\n',
    'author': 'Amos Jun-yeung Ng',
    'author_email': 'me@amos.ng',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Module for defining a single iteration of the MRKL agent loop."""
from __future__ import annotations

from typing import Any, List, Type

from langchain.schema import BaseLanguageModel
from langchain.tools.base import BaseTool

from langchain_contrib.chains import ChoiceChain, ToolChain

from .pick_action import MrklPickActionChain
from .prompt import MrklPromptSelector


class MrklLoopChain(ChoiceChain):
    """Chain executing one single iteration of the MRKL agent."""

    @classmethod
    def from_tools(
        cls,
        llm: BaseLanguageModel,
        tools: List[BaseTool],
        prompt_selector: Type[MrklPromptSelector] = MrklPromptSelector,
        **kwargs: Any,
    ) -> MrklLoopChain:
        """Create a new instance of the chain from tools."""
        picker = MrklPickActionChain.from_tools(llm, tools, prompt_selector)
        choices = {tool.name: ToolChain(tool=tool) for tool in tools}
        return cls(
            choice_picker=picker, choices=choices, ignore_keys=["observation"], **kwargs
        )



"""Make Terminal available in langchain Tool form."""

from langchain.tools.base import BaseTool
from pydantic import Field

from .terminal import Terminal


class TerminalTool(BaseTool):
    """Access to the Terminal in langchain's default Tool form."""

    name: str = "Terminal"
    description: str = (
        "Executes commands in a terminal. Input should be valid commands, and the "
        "output will be any output from running that command."
    )
    terminal: Terminal = Field(default_factory=Terminal)

    def _run(self, tool_input: str) -> str:
        """Use the terminal."""
        return self.terminal.run_bash_command(tool_input)

    async def _arun(self, tool_input: str) -> str:
        """Use the terminal asynchronously."""
        return self._run(tool_input)



"""Module that defines human-based 'LLM's."""

import logging
from typing import List, Optional

from langchain.llms.base import LLM
from langchain.llms.utils import enforce_stop_tokens
from simple_term_menu import TerminalMenu

from langchain_contrib.prompts.choice import ChoiceStr


class BaseHuman(LLM):
    """Base class for humans acting as LLMs."""

    @property
    def _llm_type(self) -> str:
        return "Human"

    def _has_stop(self, input: str, stop: List[str]) -> bool:
        """Whether or not the input contains any of the current stops."""
        for s in stop:
            if s in input:
                logging.debug(f"Found '{repr(s)}' in {input}")
                return True
        return False

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        """Get the human to respond to the given prompt."""
        if stop is not None:
            assert isinstance(stop, list), "Stops must be provided as a list"
        user_input = input(prompt)

        if stop is None or stop == ["\n"]:
            return user_input

        while not self._has_stop(user_input, stop):
            user_input += "\n" + input()

        return enforce_stop_tokens(user_input, stop)


class Human(BaseHuman):
    """A human 'LLM' with terminal support for choice prompts."""

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        """Get the human to respond to the given prompt."""
        if isinstance(prompt, ChoiceStr):
            try:
                print(prompt)
                print()
                result = TerminalMenu(prompt.choices).show()
                if result is None:
                    raise ValueError("No menu response; user might've hit ctrl-C")
                return prompt.choices[result]
            except OSError:
                # likely because of: [Errno 6] No such device or address: '/dev/tty'
                return super()._call("", stop)

        return super()._call(prompt, stop)



"""Defines the Prefixed prompt template type."""
from __future__ import annotations

from typing import Optional

from langchain.prompts.base import BasePromptTemplate
from pydantic import BaseModel, Extra

from .chained import ChainedPromptTemplate
from .schema import Templatable, into_template


class PrefixedTemplate(BaseModel):
    """Wraps another prompt template into one that can take in a prefix.

    This is useful for when you want to add a prefix to a prompt,
    but you don't want to have to do it manually.
    """

    template: BasePromptTemplate

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    def __init__(self, templatable: Templatable) -> None:
        """Initialize the PrefixedTemplate.

        Args:
            templatable: The template to wrap.
        """
        super().__init__(template=into_template(templatable))

    def __call__(
        self, prefix: Optional[Templatable] = None, joiner: str = ""
    ) -> BasePromptTemplate:
        """Create the final PromptTemplate with a prefix.

        Args:
            prefix: The prefix to add to the existing template.
            joiner: The string to put in between the prefix and the template.

        Returns:
            The final template with the prefix added if specified.
        """
        if prefix:
            return ChainedPromptTemplate([prefix, self.template], joiner=joiner)
        return self.template



"""Module to change the current directory."""

import os
from collections.abc import Generator
from contextlib import contextmanager
from typing import Optional


@contextmanager
def current_directory(path: Optional[str] = None) -> Generator:
    """Change the current directory for the duration of this context.

    Restores the current directory back to its original value upon exiting the context,
    even if more directory changes have been made in the meantime. If no path is
    provided as input, the current directory will be saved and restored on context
    exit.
    """
    if path is None:
        path = "."

    og_cwd: str = os.getcwd()
    os.makedirs(path, exist_ok=True)
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(og_cwd)


@contextmanager
def temporary_file(path: str, check_creation: bool = True) -> Generator:
    """Ensure that a temporarily created file will stop existing on exit.

    Useful for cleanup after tests. By default, this also checks that the file will
    have been successfully created during the test.
    """
    if os.path.isfile(path):
        os.remove(path)

    try:
        yield
        assert not check_creation or os.path.isfile(path)
    finally:
        if os.path.isfile(path):
            os.remove(path)



"""Patch for langchain tooling."""

from langchain.agents.load_tools import _BASE_TOOLS

from langchain_contrib.tools.terminal.tool import TerminalTool


def _get_persistent_terminal() -> TerminalTool:
    return TerminalTool()


_BASE_TOOLS["persistent_terminal"] = _get_persistent_terminal



"""Module to join F-strings."""

from typing import List, Union

from fvalues import F


def f_join(joiner: str, substrings: List[Union[str, F]]) -> F:
    """Join strings together while preserving their original F and non-F status.

    This function exists to temporarily provide this functionality pending the merge of
    https://github.com/oughtinc/fvalues/pull/11

    Args:
        joiner: The string to join the substrings with.
        substrings: The substrings to join. Can be either regular strings or F-strings.

    Returns:
        The joined string, with all original parts preserved.
    """
    if substrings == []:
        return F("")

    joined = joiner.join(substrings)
    parts = []
    for substring in substrings:
        parts.append(substring)

        # if it's the empty string, we can just avoid polluting parts with it
        if joiner != "":
            parts.append(joiner)

    if joiner != "":  # joiner should only exist in between parts
        parts.pop()

    return F(joined, parts=tuple(parts))



"""Module to remove ANSI escape sequences from strings."""

import re
from typing import List


def ansi_escape_regex() -> re.Pattern:
    """Return a regex pattern to match ANSI escape sequences."""
    # not really ANSI escape sequence, but similar vibe
    rs = r"\r|\\r"

    # how we know we're in an ANSI escape sequence
    escape_prefixes = [r"\\033", r"\\e", r"\x1b"]
    brackets = [r"\[", r"\("]
    all_ansi_escapes = [
        prefix + bracket for prefix in escape_prefixes for bracket in brackets
    ]
    all_escapes_regex = "|".join(all_ansi_escapes)

    # the actual command once we know there's an escape
    escape_command = r"[\d;]*m|\d*[A-K]"

    escapes = f"({all_escapes_regex})({escape_command})"
    return re.compile("|".join([rs, escapes]))


def remove_match(line: str, next_match: re.Match) -> str:
    """Remove the next match from the line."""
    return line[: next_match.start()] + line[next_match.end() :]


def remove_ansi_escapes(input: str) -> str:
    r"""Remove ANSI escape sequences from the input string.

    Assumes that all \r\n have already been replaced by \n.

    Incomplete escape interpreter based on
    https://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html

    Additional documentation available at
    https://en.wikipedia.org/wiki/ANSI_escape_code#CSI_(Control_Sequence_Introducer)_sequences
    """
    regex = ansi_escape_regex()
    cleaned: List[str] = []
    for line in input.split("\n"):
        next_match = re.search(regex, line)
        while next_match is not None:
            full_match = next_match.group()
            _, command = next_match.groups()
            if full_match == "\r" or full_match == "\\r":
                line = line[next_match.end() :]
            elif command.endswith("A"):
                num_lines = int(command[:-1])
                cleaned = cleaned[:-num_lines]
                line = line[next_match.end() :]
            elif command.endswith("m"):
                line = remove_match(line, next_match)
            elif command.endswith("J"):
                if command == "0J":
                    # todo: track cursor position properly
                    # do nothing because it only affects things after the cursor
                    pass
                line = remove_match(line, next_match)
            elif command.endswith("K"):
                if command == "K" or command == "0K":
                    # erase after cursor. no-op because not tracking cursor
                    line = remove_match(line, next_match)
                else:
                    line = line[next_match.end() :]
            else:  # just remove the ANSI escape code if we can't interpret it
                line = remove_match(line, next_match)
            next_match = re.search(regex, line)
        cleaned.append(line)
    return "\n".join(cleaned)


def interpret_terminal_output(input: str) -> str:
    """Render approximately how terminal output looks on screen."""
    return remove_ansi_escapes(input.replace("\r\n", "\n"))
